'***************************************************************************************
'
' MikronautsSVGA50
'
' Copyright 2011 William Henning
'
' http://Mikronauts.com
' mikronauts@gmail.com
'
' DESCRIPTION:
'
' SINGLE COG 50 column text mode driver with varying number of rows of text
'
'       50x75, 50x60, 50x50, 50x40, 50x30, 50x25, 50x20
'       50x15, 50x12, 50x10, 50x8,  50x6,  50x5
'
'
' THEORY OF OPERATION:
'
' - Generate hsync/vsync timing same as SVGA 800x600 @ 60Hz
' - use 20MHz dot clock and scale active video, hf,hs,hb by 0.5 so timing so monitor
'   'sees' same timing as full VGA with 25Mhz dot clock would have
' - I am using 20MHz clock and staying in sync with the hub and I still had to
'   unroll to make it possible
' - this way there is time, since I got tricky, to display 50 columns of text
' - got tricky with font interleaving so two hub ops enough per byte
' - separate foreground/background colors for each line of text
' - "scanx" allows setting vertical scaling for fonts
' - five fonts of varying heights, allows for a total of 13 different text modes from
'   this one driver
' - NOTE: hf,hs,hb, vf,vs,vb may need to be tuned to specific monitors for this to work
'   hassle-free for diferent monitors
'
' NOTES:
'
' I have two other drivers that use these same fonts to provide many more text modes:
'
' - MikronautsVGA80 provides 80 colums; requires the Prop to run at 100Mhz (16*6.25MHz)
' - MikronautsVGA64 provides 50 columns; runs at 16*5.00Mhz
' - MikronautsVGA40 provides 40 columns; runs at 16*5.00Mhz
'
' TERMS OF USE: MIT License
'
' Permission is hereby granted, free of charge, to any person obtaining a copy
' of this software and associated documentation files (the "Software"), to deal
' in the Software without restriction, including without limitation the rights
' to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
' copies of the Software, and to permit persons to whom the Software is
' furnished to do so, subject to the following conditions:
'
' The above copyright notice and this permission notice shall be included in
' all copies or substantial portions of the Software.
'
' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
' IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
' FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
' AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
' LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
' OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
' THE SOFTWARE.
'
' CREDITS:
'
' Chip Gracey for the Propeller and multi-cog hirez text object, from which I adopted
' the initialization code, constant names, and hsync/vsync code
'
' RELEASE INFORMATION:
'
' Feb.19/2011 - v0.80 initial beta release
' Feb.20/2011 - v0.81 added VGA_Text compatibility, ScreenAddr, ColorAddr and FontAddr
' Apr.03/2011 - v1.00 many thanks to kuroneko for his brilliant forcing hub sync code patch!
' Apr.03/2011 - v1.01 added improved sync patch from kuroneko
'
' I also wish to thank everyone who contributed to these driver threads
'
'***************************************************************************************


OBJ

  font: "mikronauts8x8font"    '1:50x75, 3:50x25, 5:50x15
' font: "mikronauts8x10font"   '1:50x60, 2:50x30, 3:50x20, 4:50x15, 5:50x12, 6:50x10 10:50x6
' font: "mikronauts8x12font"   '1:50x50, 2:50x25, 5:50x10, 10:50x5
' font: "mikronauts8x15font"   '1:50x40, 2:50x20, 4:50x10, 5:50x8, 8:50x5

'50x75, 50x60, 50x50, 50x40, 50x30, 50x25, 50x20, 50x15, 50x12, 50x10, 50x8, 50x6, 50x5

CON

  scanx   = 2
  height  = font#height

' VGA timing taken from http://tinyvga.com/vga-timing/640x480@60Hz

  hp = 512        'horizontal pixels, scaled by .8

  hf = 8'12'8       'horizontal front porch pixels, scaled by .8
  hs = 40'76'40      'horizontal sync pixels, scaled by .8
  hb = 80'40'80      'horizontal back porch pixels, scaled by .8

  vp = 480/scanx  'vertical pixels

  vf = 10         'vertical front porch lines - orig 9
  vs = 2          'vertical sync lines
  vb = 33         'vertical back porch lines - orig 28

  hn = 1          'horizontal normal sync state (0|1)
  vn = 1          'vertical normal sync state (0|1)

  pll_x = 4       'must use 4 for pseudo-synced video and hub access

' Screen Specs

  chars = hp / 8
  lines = vp / height

' VGA_Text method support

  lastrow = chars * lines - chars

  rows = lines
  cols = chars

  screensize = cols * rows

VAR

  long cog, row, col, flag
  long mailbox[4]

  long colors[lines]
  byte tilemap[chars*lines]

PUB start(BasePin) : okay
  'if driver is already running, stop it
  stop

  'implant pin settings and pointers, then launch COG
  pixel_base:=@tilemap
  color_base:=@colors
  font_base:=font.addr
  reg_vcfg := $200000FF + (BasePin & %011000) << 6
  reg_dira := $FF << (BasePin & %011000)

  mailbox[0]:=@tilemap
  mailbox[1]:=@colors
  mailbox[2]:=$200000FF + (BasePin & %011000) << 6
  mailbox[3]:=$FF << (BasePin & %011000)

  col := 0
  row := 0

  if (cog := cognew(@init,@mailbox[0]) + 1)
    return true

PUB ScreenAddr
  return @tilemap

PUB ColorAddr
  return @colors

PUB FontAddr
  return font.addr

PUB SetColor(n,v)
  colors[n]:=v

PUB put(xc,yc,ch)
  tilemap[yc*chars + xc] := ch

PUB cls|yn
  'init colors to cyan on blue
  repeat yn from 0 to lines-1
    colors[yn]:=$2804
  ' fill the screen with space character
  bytefill(@tilemap,32,lines*chars)

PUB scroll|k
  k := tilemap[0]
  bytemove(@tilemap,@tilemap+1,chars*lines-1)
  tilemap[chars*lines-1]:=k


PUB stop | i
  if cog
    cogstop(cog~ - 1)


'---------------- VGA_Text Routines -----------------------

PUB str(stringptr)

'' Print a zero-terminated string

  repeat strsize(stringptr)
    out(byte[stringptr++])


PUB dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    out("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10


PUB hex(value, digits)

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))


PUB bin(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")


PUB out(c) | i, k

'' Output a character
''
''     $00 = clear tilemap
''     $01 = home
''     $08 = backspace
''     $09 = tab (8 spaces per)
''     $0A = set X position (X follows)
''     $0B = set Y position (Y follows)
''     $0C = not supported - set color (color follows)
''     $0D = return
''  others = printable characters

  case flag
    $00: case c
           $00: bytefill(@tilemap, 32, screensize)
                col := row := 0
           $01: col := row := 0
           $08: if col
                  col--
           $09: repeat
                  print(" ")
                while col & 7
           $0A..$0C: flag := c
                     return
           $0D: newline
           other: print(c)
    $0A: col := c // cols
    $0B: row := c // rows
    $0C: ' do nothing
  flag := 0


'PUB setcolors(colorptr) is not supported, see SetColor()


PRI print(c)

  tilemap[row * cols + col] := c
  if ++col == cols
    newline


PRI newline | i

  col := 0
  if ++row == rows
    row--
    bytemove(@tilemap, @tilemap[cols], lastrow)   'scroll lines
    bytefill(@tilemap[lastrow], 32, cols)      'clear new line

'---------------- End of VGA_Text Routines -----------------------

'' New functions

PUB get(xc, yc)

  return tilemap[yc*chars+xc]


PUB blink(t{ms}) | ch

  ch:= get(col, row)
  waitcnt (cnt+clkfreq/1000*t)
  put(col, row, 95)
  waitcnt (cnt+clkfreq/1000*t)
  put(col, row, ch)




DAT

                        org                             'set origin to $000 for start of program

init                    'mov     x,par
                        'rdlong  pixel_base,par
                        'add     x,#4
                        'rdlong  color_base,x
                        'add     x,#4
                        'rdlong  reg_vcfg,x
                        'add     x,#4
                        'rdlong  reg_dira,x

                        mov     dira,reg_dira           'set pin directions
                        movi    ctra,#%00001_101        'enable PLL in ctra
                        movi    frqa,#pll_x << 3        'set pixel rate
                        mov     vcfg,reg_vcfg           'set video configuration

                        ' give time for the video pll to settle

                        mov     vscl,hvis
                        waitvid hv,#0

' Main loop, display field and do invisible sync lines
                          
field                   mov     color_ptr,color_base    'reset color pointer
                        mov     screen_ptr,pixel_base   'reset pixel pointer
                        mov     y,#lines                'set y lines of chars

'========================================================================================
:ytile                  rdlong  color,color_ptr         'get color word
                        or      color,hv                'set h/v inactive states
                        mov     yl,#height              'set y lines per tile
                        mov     font_ptr,font_base

:yline                  mov     yx,#scanx               'set y expansion

:yexpand                mov     x,#chars                'set x tiles
                        mov     vscl,vscl_pixel         'set pixel vscl
                        mov     ccnt, #chars-1

                        '------------------------------------- 00
                        rdbyte  tile,screen_ptr
                        add     screen_ptr,#1
                        add     tile,font_ptr
                        rdbyte  pixel,tile
                        waitvid color,pixel
                        '------------------------------------- 01-63
                        rdbyte  tile,screen_ptr
                        add     screen_ptr,#1
                        add     tile,font_ptr
                        rdbyte  pixel,tile
                        cmp     color,pixel
                        djnz    ccnt, #$-5

                        sub     screen_ptr,#chars       'repoint to first pixels in same line
                        mov     x,#1                    'do horizontal sync
                        call    #hsync

                        djnz    yx,#:yexpand            'y expand?

                        add     font_ptr,#256

                        djnz    yl,#:yline              'another y line in same tile?
                        add     screen_ptr,#chars       'point to first pixels in next line

                        sub     font_ptr,font_table_size

                        add     color_ptr,#4            'point to next color word
                        djnz    y,#:ytile               'another y tile?

'========================================================================================

                        mov     x,#vf                   'do vertical front porch lines
                        call    #blank
                        mov     x,#vs                   'do vertical sync lines
                        call    #vsync
                        mov     x,#vb                   'do vertical back porch lines
                        call    #vsync

                        jmp     #field                  'field done, loop
                        

' Subroutine - do blank lines

vsync                   xor     hvsync,#$101            'flip vertical sync bits

blank                   mov     vscl,hvis               'do blank pixels
                        waitvid hvsync,#0

hsync                   mov     vscl,#hf                'do horizontal front porch pixels
                        waitvid hvsync,#0
                        mov     vscl,#hs                'do horizontal sync pixels
                        waitvid hvsync,#1

                        neg     href, cnt               ' waitvid reference ...
                        cogid   $ nr
                        add     href, cnt               ' ... relative to hub window
                        sub     href, #(hs + hb) * 4    ' pixels start in (hs + hb)*4
                        shr     href, #2                ' delta in frame clocks
                        and     href, #%111             ' limit to 0..7
                        add     href, #hb               ' add base -> hb + 0..7
                        sub     href, #6                ' apply fixed shift
                        
                        mov     vscl,href               'do horizontal back porch pixels
                        waitvid hvsync,#0
                        djnz    x,#blank                'another line?
hsync_ret
blank_ret
vsync_ret               ret

' Data

reg_dira                long    0                       'set at runtime
reg_vcfg                long    0                       'set at runtime
color_base              long    0                       'set at runtime (2 contiguous longs)
pixel_base              long    0                       'set at runtime
font_base               long    0

vscl_pixel              long    1 << 12 + 8            '1 pixel per clock and 32 pixels per set
hvis                    long    hp                      'visible pixels per scan line
hv                      long    ((hn << 1 + vn) * $0101) 'hv_inactive             '-H,-V states
hvsync                  long    ((hn << 1 + vn) * $0101) ^ $200      '+/-H,-V states

tile                    long    0
trow                    long    0
font_table_size         long    height * 256
color_ptr               long    0
screen_ptr              long    0
font_ptr                long    0
color                   long    0
pixel                   long    0
x                       long    0
y                       long    0
yl                      long    0
yx                      long    0

href                    res     1
ccnt                    res     1

                        fit
                        
DAT